//! # Tools for splitting strings into words
//!
//! Tokenizers just provide handy methods to split a string into words,
//! following certain rules.

mod main;

pub use self::main::split_whitespace_remove_punc;
